/*
 *
 * Copyright (C) 2010-2012 Fonpit AG
 *
 */
package de.androidpit.licensing;

import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;
import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.Signature;
import java.security.SignatureException;
import java.security.spec.RSAPublicKeySpec;
import java.util.Random;

import android.util.Log;

import com.google.android.vending.licensing.util.Base64;
import com.google.android.vending.licensing.util.Base64DecoderException;

import de.androidpit.app.services.SignedResponse;

/**
 * The implementation of the license validator which checks the validity of the
 * license response in-app.
 * 
 * @author <a href="mailto:carsten.rose@androidpit.de">Carsten W. Rose</a>
 */
final class AndroidPitLicensingResponseValidator
{

    private final int mSalt;
    private final String mPackageName;

    private static final String SIGNATURE_ALGORITHM = "SHA1withRSA";

    private static final String LOG_TAG = "AndroidPitLicensingResponseValidator";

    /**
     * Creates the license response validator.
     * 
     * @param packageName
     *            the app's package name
     */
    AndroidPitLicensingResponseValidator(String packageName)
    {
        mSalt = new Random().nextInt();
        mPackageName = packageName;
    }

    /**
     * A salt to make the signed data more unpredictable, e.g. a random number.
     * Do not forget to remember thes salt from the request at least until you
     * have received the response and validated the salt.
     * 
     * @return the salt as an integer
     */
    public int getSalt()
    {
        return mSalt;
    }

    /**
     * Checks the response on validity if it comes from the server and extract
     * the response code. If the validation of the signed repsonse failes. It
     * has to return {@link AndroidPitLicenseCheckCode}.NOT_LICENSED
     * 
     * @see AndroidPitResponseData for the expected structure of the signedData
     *      field in SignedResponse
     * 
     * @param publicKey
     *            the public key of the developer, provided by the app.
     * @param response
     *            the server response forwarded by the AppCenter.
     * 
     * @return the response code of the AppCenter.
     */
    public AndroidPitLicenseCheckCode checkResponse(
            String publicKeyStr,
            SignedResponse response)
    {
        AndroidPitLicenseCheckCode responseCode = AndroidPitLicenseCheckCode
                .valueOf(response.getIsLicensed());

        // only check on positive license message
        if (responseCode.equals(AndroidPitLicenseCheckCode.LICENSED))
        {
            try
            {
                PublicKey pubKey = createPublicKey(publicKeyStr);
                Signature sig = Signature.getInstance(SIGNATURE_ALGORITHM);
                sig.initVerify(pubKey);
                sig.update(response.getSignedData().getBytes());

                if (!sig.verify(Base64.decode(response.getSign())))
                {
                    Log.e(LOG_TAG, "Failed to verify signature of response");
                    return AndroidPitLicenseCheckCode.NOT_LICENSED;
                }
            }
            catch (NoSuchAlgorithmException ex)
            {
                Log.e(LOG_TAG, "No such algorithm", ex);
                throw new RuntimeException(ex);
            }
            catch (InvalidKeyException ex)
            {
                Log.e(LOG_TAG, "Invalid public key");
                return AndroidPitLicenseCheckCode.ERROR_INVALID_PUBLIC_KEY;
            }
            catch (SignatureException ex)
            {
                Log.e(LOG_TAG, "Bad signature");
                throw new RuntimeException(ex);
            }
            catch (Base64DecoderException ex)
            {
                Log.e(
                        LOG_TAG,
                        "Failed to decode base64 string of signature",
                        ex);
                return AndroidPitLicenseCheckCode.NOT_LICENSED;
            }

            AndroidPitResponseData responseData = AndroidPitResponseData
                    .parse(response.getSignedData());

            if (responseData.verifiesAgainst(
                    mPackageName,
                    response.getIsLicensed(),
                    mSalt))
            {
                Log.i(LOG_TAG, "Verification successful");
                return responseCode;
            }
            else
            {
                Log.e(LOG_TAG, "Signature verification failed - invalid data");
                return AndroidPitLicenseCheckCode.NOT_LICENSED;
            }
        }
        else
        {
            // return App Center message
            return responseCode;
        }
    }

    /**
     * Converts the base64 encoded representation of a public key into a
     * PublicKey object.
     */
    private PublicKey createPublicKey(String pubKeyBase64)
    {
        try
        {
            // the pub key comes in as a Base64 coded string. Decode to the
            // byte array which contains the object stream of the public key
            ByteArrayInputStream pubKeyByteArray = new ByteArrayInputStream(
                    Base64.decode(pubKeyBase64));
            ObjectInputStream publicKeyObject = new ObjectInputStream(
                    pubKeyByteArray);
            BigInteger modulus = (BigInteger) publicKeyObject.readObject();
            BigInteger exponent = (BigInteger) publicKeyObject.readObject();

            RSAPublicKeySpec keySpec = new RSAPublicKeySpec(modulus, exponent);
            KeyFactory rsaKeyFactory = KeyFactory.getInstance("RSA");
            return rsaKeyFactory.generatePublic(keySpec);
        }
        catch (Exception ex)
        {
            Log.e(LOG_TAG, "Deserialization of public key failed", ex);
            return null;
        }
    }

}
